import logging

from .conf import config
from .router import router
from .rules import front_door_check_access
from .utils import get_client_ip

logger = logging.getLogger(__name__)


class FrontDoorMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response
        if config.ENABLED:
            logger.debug(f"FrontDoor enabled. Policy: {config.DEFAULT_POLICY_LABEL}")
        else:
            logger.debug("FrontDoor disabled")

    def __call__(self, request):
        ip = get_client_ip(request)
        request.user_ip = ip
        if front_door_check_access(request):
            response = self.get_response(request)
        else:
            response = router.route(request)
        return response



import logging

from django.conf import settings

from .conf import LOG_FAIL, LOG_PASS, LOG_RULE_FAIL, LOG_RULE_PASS, config
from .exception import FrontDoorAccessDenied
from .utils import get_client_ip

logger = logging.getLogger(__name__)


def allowed_ip(request, ip, **kwargs):
    return ip in config.ALLOWED_IPS


def internal_ip(request, ip, **kwargs):
    return ip in settings.INTERNAL_IPS


def special_header(request, ip, **kwargs):
    if config.HEADER and config.TOKEN:
        return request.headers.get(config.HEADER) == config.TOKEN


def has_header(request, ip, **kwargs):
    if config.HEADER:
        return config.HEADER in request.headers


def allowed_path(request, ip, **kwargs):
    return request.path in config.ALLOWED_PATHS


def forbidden_path(request, ip, **kwargs):
    if request.path in config.FORBIDDEN_PATHS:
        raise FrontDoorAccessDenied


def cookie_exists(request, ip, **kwargs):
    return config.COOKIE_NAME in request.COOKIES


def cookie_value(request, ip, **kwargs):
    if not config.COOKIE_PATTERN and config.COOKIE_NAME:
        return False
    try:
        value = request.COOKIES[str(config.COOKIE_NAME)]
    except KeyError:
        return False
    return bool(config.COOKIE_PATTERN.match(value))


def front_door_check_access(request):
    if not config.ENABLED:
        access_allowed = True
    else:
        access_allowed = config.DEFAULT_POLICY
        ip = get_client_ip(request)
        path = request.path
        extra = {"ip": ip,
                 "path": path}
        for check in config.rules:
            try:
                if check(request, ip):
                    if config.LOG_LEVEL & LOG_RULE_PASS:
                        logger.error(f"{check.__name__}() PASS", extra=extra)
                    access_allowed = True
                    break
                else:
                    if config.LOG_LEVEL & LOG_RULE_FAIL:
                        logger.error(f"{check.__name__}() FAIL", extra=extra)
            except FrontDoorAccessDenied:
                if config.LOG_LEVEL & LOG_RULE_FAIL:
                    logger.error(f"{check.__name__}() DENY", extra=extra)
                access_allowed = False
                break
        if access_allowed and (config.LOG_LEVEL & LOG_PASS):
            logger.error("Access ALLOWED", extra=extra)
        elif config.LOG_LEVEL & LOG_FAIL:
            logger.error("Access DENIED", extra=extra)

    return access_allowed



from django import forms
from django.utils.translation import gettext as _

# class FrontDoorLogLevel(forms.CharField):
from front_door.conf import LOG_NONE, LOG_FAIL, LOG_PASS, LOG_RULE_FAIL, LOG_RULE_PASS, LOG_ALL


class FrontDoorLogLevel(forms.ChoiceField):
    def __init__(self, **kwargs):
        choices = ([LOG_NONE, _("None")],
                   [LOG_FAIL, "Log Access Failures"],
                   [LOG_PASS, "Log Access"],
                   [LOG_RULE_FAIL, "Log Rule Failure"],
                   [LOG_RULE_PASS, "Log Rule Pass"],
                   [LOG_ALL, "Log Everything"],
                   )
        super().__init__(choices=choices, **kwargs)

    def clean(self, value):
        # value = super().clean(value)
        return int(value)



from django.http import HttpResponse, HttpResponseRedirect

from .conf import config


class DefaultRouter:
    def route(self, request):
        if config.REDIR_URL:
            response = HttpResponseRedirect(config.REDIR_URL)
        else:
            response = HttpResponse("", status=config.ERROR_CODE)
        return response


router = DefaultRouter()



import logging
import os
import re
from itertools import chain

from django.conf import settings
from django.utils.functional import SimpleLazyObject, cached_property
from django.utils.module_loading import import_string
from django_regex.utils import RegexList
from front_door.utils import parse_bool, parse_policy

FRONT_DOOR_CONFIG = getattr(settings, "FRONT_DOOR_CONFIG", "front_door.conf.DjangoSettings")

logger = logging.getLogger(__name__)

FORBID = False
ALLOW = True

_POLICIES = {
    FORBID: "FORBID",
    ALLOW: "ALLOW"
}
LOG_NONE = 0  # do not log
LOG_FAIL = 1  # log success
LOG_PASS = 2  # log
LOG_RULE_FAIL = 4
LOG_RULE_PASS = 8
LOG_ALL = 255


class FrontDoorConfig:
    defaults = dict(
        ALLOWED_IPS=[],  # allowed ips
        ALLOWED_PATHS=[],  # url paths regex list always allowed
        COOKIE_NAME=None,
        COOKIE_PATTERN=None,
        DEFAULT_POLICY=FORBID,
        LOG_LEVEL=LOG_NONE,
        ENABLED=False,  # FrontDoor enable/disable
        ERROR_CODE=404,  # status code if access denied
        FORBIDDEN_PATHS=[],  # url paths regex list always denied
        HEADER=None,  # special header name without HTTP- prefix
        REDIR_URL="",  # HttpResponseRedirect(REDIR_URL) if access denied
        ROUTER="front_door.router.DefaultRouter",
        RULES=[
            "front_door.rules.internal_ip",  # grant access to settings.INTERNAL_IPS
            "front_door.rules.forbidden_path",  # DENY access to FORBIDDEN_PATHS
            "front_door.rules.allowed_ip",  # grant access to FORBIDDEN_PATHS
            "front_door.rules.allowed_path",  # grant access to ALLOWED_PATHS
            "front_door.rules.special_header",  # grant access if request has Header[HEADER] == TOKEN
            "front_door.rules.has_header",  # grant access if request has HEADER
            "front_door.rules.cookie_value",  # grant access if request.COOKIES[COOKIE_NAME]
            "front_door.rules.cookie_exists",  # grant access ir COOKIE_NAME in request.COOKIES
        ],
        TOKEN=None,  # custom header value
    )

    def __init__(self) -> None:
        super().__init__()
        from django.core.signals import setting_changed
        setting_changed.connect(self.on_setting_changed)

    def _get(self, key):
        return self.defaults.get(key, None)

    def _get_list(self, key):
        custom = self._get(key)
        if isinstance(custom, (list, tuple)):
            return custom
        elif isinstance(custom, str):
            return custom.split(";")
        else:
            return []

    def __getattr__(self, key):
        if key in self.defaults.keys():
            return self._get(key)
        raise AttributeError(key)

    @cached_property
    def LOG_LEVEL(self):
        return int(self._get("LOG_LEVEL") or "0")

    @cached_property
    def DEFAULT_POLICY(self):
        return parse_policy(self._get("DEFAULT_POLICY"))

    @cached_property
    def DEFAULT_POLICY_LABEL(self):
        return _POLICIES[parse_policy(self._get("DEFAULT_POLICY"))]

    @cached_property
    def COOKIE_PATTERN(self):
        try:
            return re.compile(self._get("COOKIE_PATTERN"))
        except (TypeError, re.error):
            return None

    @property
    def ALLOWED_PATHS(self):
        rules = RegexList([])
        for rule in self._get_list('ALLOWED_PATHS'):
            if rule:
                try:
                    rules.append(rule)
                except Exception as e:
                    logger.exception(e)
        return rules

    @cached_property
    def FORBIDDEN_PATHS(self):
        rules = RegexList([])
        for rule in self._get_list('FORBIDDEN_PATHS'):
            try:
                rules.append(rule)
            except Exception as e:
                logger.exception(e)
        return rules

    @cached_property
    def ALLOWED_IPS(self):
        return self._get_list('ALLOWED_IPS')

    @cached_property
    def ERROR_CODE(self):
        return int(self._get('ERROR_CODE'))

    @cached_property
    def ENABLED(self):
        return parse_bool(self._get('ENABLED'))

    @cached_property
    def rules(self):
        result = []
        for x in self._get_list('RULES'):
            try:
                if isinstance(x, str):
                    result.append(import_string(x))
                else:
                    result.append(x)
            except Exception as e:
                logger.exception(e)
        return result

    def on_setting_changed(self, **kwargs):
        self.invalidate()

    def invalidate(self):
        for attr in chain(self.defaults.keys(), ["rules"]):
            try:
                delattr(self, attr)
            except AttributeError:
                pass


class DjangoSettings(FrontDoorConfig):
    def _get(self, key):
        full_name = f"FRONT_DOOR_{key}"
        return getattr(settings, full_name, super()._get(key))


class DjangoConstance(DjangoSettings):

    def __init__(self) -> None:
        import constance.signals
        super().__init__()

        constance.signals.config_updated.connect(self.on_setting_changed)
        self.storage = constance.config

    def _get(self, key):
        full_name = f"FRONT_DOOR_{key}"
        return getattr(self.storage, full_name, super()._get(key))


class OSEnv(FrontDoorConfig):
    def _get(self, key):
        full_name = f"FRONT_DOOR_{key}"
        return os.environ.get(full_name, self.defaults.get(key, None))


def get_config():
    return import_string(FRONT_DOOR_CONFIG)()


config = SimpleLazyObject(get_config)

#
# def get_front_door_allowed_ips():
#     return config.ALLOWED_IPS
#
#
# def get_front_door_internal_ips():
#     return settings.INTERNAL_IPS
#
#
# def get_front_door_whitelisted_paths():
#     return config.ALLOWED_PATHS
#
#
# def get_front_door_forbidden_paths():
#     return config.FORBIDDEN_PATHS
#
#
# def front_door_is_active():
#     return config.ENABLED



import logging

from django.http import HttpRequest

logger = logging.getLogger(__name__)


def get_client_ip(request: HttpRequest):
    """
    type: (WSGIRequest) -> Optional[Any]
    Naively yank the first IP address in an X-Forwarded-For header
    and assume this is correct.

    Note: Don't use this in security sensitive situations since this
    value may be forged from a client.
    """
    for x in [
        "HTTP_X_ORIGINAL_FORWARDED_FOR",
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_REAL_IP",
        "REMOTE_ADDR",
    ]:
        ip = request.META.get(x)
        if ip:
            if "," in ip:
                return ip.split(",")[0].strip()
            return ip


def parse_bool(value):
    return value in [True, "True", "true", "T", "1", 1, "y", "Y"]


def parse_policy(value):
    return parse_bool(value) or (str(value).lower() in ["allow", "open"])



#!/usr/bin/env python
import ast
import os
import re
from setuptools import find_packages, setup

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
init = os.path.join(ROOT, 'src', 'front_door', '__init__.py')

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open(init, 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

requirements = ["django_regex"]
constance_require = ["django-picklefield", "django-constance"]
redis_require = ["redis", ]
tests_require = ["django-webtest",
                 "isort",
                 "flake8",
                 "pytest",
                 "pytest-django",
                 "pytest-echo",
                 "tox",
                 "django-picklefield",
                 "django-constance",
                 "redis",
                 "pytest-coverage"]
ua_require = ["user_agents",]
dev_require = ["pdbpp",
               "django"]
docs_require = []

setup(
    name='django-front-door',
    version=version,
    url='https://github.com/saxix/django-front-door',
    download_url='https://github.com/saxix/django-front-door',
    author='sax',
    author_email='s.apostolico@gmail.com',
    description="",
    license='MIT',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    install_requires=requirements,
    tests_require=tests_require,
    extras_require={
        'test': requirements + tests_require + constance_require + redis_require + ua_require,
        'dev': dev_require + tests_require,
        'docs': dev_require + docs_require,
        'constance': constance_require,
        'redis': redis_require,
        'ua': ua_require
    },
    zip_safe=False,
    platforms=['any'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Operating System :: OS Independent',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers'],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)
